# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import importlib

from typing import Any


def import_class(class_: str) -> Any:
    """Imports a class described in the given String"""
    class_parts = class_.split('.')
    class_name = class_parts.pop()
    module_name = '.'.join(class_parts)

    if module_name not in sys.modules:
        importlib.import_module(module_name)
    Class = getattr(sys.modules.get(module_name), class_name)

    return Class



from typing import List

from pydantic import parse_obj_as, TypeAdapter

from orchd_sdk.models import SensorTemplate, Ref, Sensor

SENSOR_TEMPLATE_BASE_ROUTE = '/sensor_template/'
SENSORS_BASE_ROUTE = '/sensor/'


class SensorClient:

    def __init__(self, orchd_client):
        self.orchd_client = orchd_client

    async def get_sensor_templates(self):
        adapter = TypeAdapter(List[SensorTemplate])
        templates = await self.orchd_client.get(SENSOR_TEMPLATE_BASE_ROUTE)
        return adapter.validate_python(templates)

    async def get_sensor_template(self, template_id: str) -> SensorTemplate:
        template = await self.orchd_client.get(
            f'{SENSOR_TEMPLATE_BASE_ROUTE}{template_id}/')
        return SensorTemplate(**template)

    async def add_sensor_template(self,
                                  template: SensorTemplate) -> SensorTemplate:
        response = await self.orchd_client.post(SENSOR_TEMPLATE_BASE_ROUTE,
                                                template.model_dump())
        return SensorTemplate(**response)

    async def remove_sensor_template(self, template_id: str) -> str:
        response = await self.orchd_client.delete(
            f'{SENSOR_TEMPLATE_BASE_ROUTE}{template_id}/')
        return response

    async def get_sensors(self):
        adapter = TypeAdapter(List[Sensor])
        sensors = await self.orchd_client.get(SENSORS_BASE_ROUTE)
        return adapter.validate_python(sensors)

    async def get_sensor(self, sensor_id: str) -> Sensor:
        sensor = await self.orchd_client.get(f'{SENSORS_BASE_ROUTE}{sensor_id}/')
        return Sensor(**sensor)

    async def add_sensor(self, template_id: str) -> Sensor:
        response = await self.orchd_client.post(
            f'{SENSORS_BASE_ROUTE}', data=Ref(id=template_id).model_dump())
        return Sensor(**response)

    async def remove_sensor(self, sensor_id: str) -> str:
        response = await self.orchd_client.delete(f'{SENSORS_BASE_ROUTE}{sensor_id}/')
        return response

    async def stop_sensor(self, sensor_id: str) -> str:
        response = await self.orchd_client.post(f'{SENSORS_BASE_ROUTE}{sensor_id}/stop', data={})
        return response

    async def start_sensor(self, sensor_id: str) -> str:
        response = await self.orchd_client.post(f'{SENSORS_BASE_ROUTE}{sensor_id}/start', data={})
        return response


from typing import Dict

from orchd_sdk.models import Event

EVENTS_BASE_ROUTE = '/events/'


class EventClient:

    def __init__(self, orch_client):
        self.orchd_client = orch_client

    async def propagate(self, event_name: str, event_data: Dict):
        await self.orchd_client.post(EVENTS_BASE_ROUTE,
                                     data=Event(event_name=event_name,
                                                data=event_data).model_dump())

    async def event_stream(self):
        return await self.orchd_client.stream(f'{EVENTS_BASE_ROUTE}event_stream')



# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re

from orchd_sdk.errors import InvalidInputError

SNAKE_CASE_REGEX = re.compile(r'\A([a-z]|(_+[a-z]))[a-z0-9_]*\Z')
KEBAB_CASE_NO_SURROUNDING_DASHES = re.compile(r'\A[a-z]([a-z0-9]|-[a-z0-9])*\Z')


def is_kebab_case(word: str):
    matched_word = KEBAB_CASE_NO_SURROUNDING_DASHES.match(word)
    return matched_word is not None


def is_snake_case(word: str):
    matched_word = SNAKE_CASE_REGEX.match(word)
    return matched_word is not None


def kebab_case_to_snake_case(kebab_cased_word: str):
    if is_kebab_case(kebab_cased_word):
        return kebab_cased_word.replace('-', '_')
    else:
        raise InvalidInputError('Given word is not a Kebab Case!')


def snake_to_camel_case(snake_cased_word: str):
    if is_snake_case(snake_cased_word):
        words = snake_cased_word.split('_')
        return ''.join(w.capitalize() for w in words)
    else:
        raise InvalidInputError('Given word is not Snake Case!')



import json

import aiohttp
from pydantic import parse_obj_as, TypeAdapter

from orchd_sdk.api.events import EventClient
from orchd_sdk.api.reactions import ReactionClient
from orchd_sdk.api.sensors import SensorClient
from orchd_sdk.api.sinks import SinkClient
from orchd_sdk.errors import handle_http_errors
from orchd_sdk.models import Event


class HTTPEventStream:
    def __init__(self, response):
        self.response = response
        self._stream = response.content

    async def __aiter__(self):
        return self

    async def __anext__(self):
        chunk = await self._stream.read(4096)
        return chunk.decode('utf-8')

    async def next(self):
        adapter = TypeAdapter(Event)
        event = json.loads(await self.__anext__())
        return adapter.validate_python(event)

    async def close(self):
        await self.response.release()


class OrchdAgentClient:

    def __init__(self, host: str, port: int, token: str = None):
        self._host = host
        self._port = port
        self._token = token
        self._session = aiohttp.ClientSession()
        if token:
            self._session.headers.update(
                {'Authorization': f'Bearer {self._token}'})

        self.reactions = ReactionClient(self)
        self.sinks = SinkClient(self)
        self.sensors = SensorClient(self)
        self.events = EventClient(self)

    def _url(self, path: str):
        return f'http://{self._host}:{self._port}/orchd/v1{path}'

    async def get(self, path: str):
        async with self._session.get(self._url(path)) as response:
            handle_http_errors(response)
            return await response.json()

    async def post(self, path: str, data: dict):
        async with self._session.post(self._url(path), json=data) as response:
            handle_http_errors(response)
            return await response.json()

    async def delete(self, path: str):
        async with self._session.delete(self._url(path)) as response:
            handle_http_errors(response)
            return await response.json()

    async def put(self, path: str, data: dict):
        async with self._session.put(self._url(path), json=data) as response:
            handle_http_errors(response)
            return await response.json()

    async def stream(self, path: str):
        response = await self._session.get(self._url(path))
        handle_http_errors(response)
        return HTTPEventStream(response)

    async def close(self):
        await self._session.close()




from typing import List

from pydantic import parse_obj_as, TypeAdapter

from orchd_sdk.models import SinkTemplate

SINK_TEMPLATES_BASE_ROUTE = '/sink_templates'


class SinkClient:
    def __init__(self, orchd_agent_client):
        self.client = orchd_agent_client

    async def add_sink_template(self, template: SinkTemplate) -> SinkTemplate:
        response = await self.client.post(SINK_TEMPLATES_BASE_ROUTE, template.model_dump())
        return SinkTemplate(**response)

    async def get_sink_templates(self):
        adapter = TypeAdapter(List[SinkTemplate])
        response = await self.client.get(SINK_TEMPLATES_BASE_ROUTE)
        return adapter.validate_python(response)

    async def get_sink_template(self, template_id: str) -> SinkTemplate:
        response = await self.client.get(f'{SINK_TEMPLATES_BASE_ROUTE}/{template_id}/')
        return SinkTemplate(**response)

    async def remove_sink_template(self, template_id: str) -> str:
        response = await self.client.delete(f'{SINK_TEMPLATES_BASE_ROUTE}/{template_id}/')
        return response



# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


class InvalidInputError(Exception):
    """
    Invalid input given.
    """


class ReactorError(Exception):
    """Raise on internal Errors in the Reactor."""


class SinkError(Exception):
    """ Raised on Sink errors."""


class ReactionError(Exception):
    """ Raised on Reaction Management and Operation errors."""


class ReactionHandlerError(Exception):
    """ Raised by ReactionHandler implementations."""


class SensorError(Exception):
    """ Raised by Sensors to indicate an not fatal error."""


class SensorFatalError(Exception):
    """ Sensor fatal error. Sensor will not be available anymore."""


class InvalidRequestError(Exception):
    """ Raised when the request is invalid."""


class NotFoundError(Exception):
    """ Raised when the resource is not found."""


class ServerError(Exception):
    """ Raised when the server is not available."""


def handle_http_errors(response):
    if response.status == 404:
        raise NotFoundError()
    elif response.status == 400 or response.status == 422:
        raise InvalidRequestError()
    elif response.status == 500:
        raise ServerError()
    elif response.status > 400:
        raise Exception(response.status, response.reason)
    return response


# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio
import uuid
from abc import ABC, abstractmethod
from asyncio import Task

import logging

from orchd_sdk.errors import SensorFatalError
from orchd_sdk.reaction import global_reactions_event_bus, ReactionsEventBus

from orchd_sdk.models import Event, SensorTemplate, Sensor

logger = logging.getLogger(__name__)

class SensorError(Exception):
    """
    Exception raised by a Sensor for errors occurring in it.
    """
    pass


class AbstractCommunicator(ABC):
    """
    A communicator establishes a link to the Reactor to send events.

    A communicator abstracts the communication between the Sensor and Reactor.
    Different Communicators can be implemented and used by the Sensors. This
    design allows the creation of new forms of communication if needed.

    Communicators need to authenticate against the orchd agent to be allowed
    to emit events on it if communicating through the network.
    """

    def __init__(self):
        self._authenticated = False
        self.id = str(uuid.uuid4())

    @abstractmethod
    async def emit_event(self, event: Event):
        """
        Emits the event in the orchd agent's Reactor.
        :param event: Event to be emitted
        """

    @abstractmethod
    def close(self):
        """
        Closes connections and releases resources.
        """

    @abstractmethod
    async def authenticate(self):
        """
        Authenticates the sensor against the Orchd Agent.
        """


class SensorState:
    READY = (2, 'READY')
    RUNNING = (3, 'RUNNING')
    STOPPED = (4, 'STOPPED')


class AbstractSensor(ABC):
    """
    Base class to be used to implement Sensors to be used in orchd.

    Sensors need to implement the logic that will detect external events
    and inject it in Orchd. To inject the event they will use the Communicator.
    """

    @abstractmethod
    def __init__(self, sensor_template: SensorTemplate,
                 communicator: AbstractCommunicator,
                 sensing_interval=0):
        self.id = str(uuid.uuid4())
        self.sensor_template = sensor_template
        self.communicator = communicator
        self.sensing_interval = sensor_template.sensing_interval or sensing_interval
        self._state = SensorState.READY
        self._start_task: Task = None
        self._events_counter = -1
        self._events_forwarded = -1
        self._events_discarded = -1

    @abstractmethod
    async def sense(self):
        """
        Sensor function to be called in order to sense an external event.

        It will be called in loop while the sensor is running.
        """

    async def _start(self):
        while self.state == SensorState.RUNNING:
            try:
                await self.sense()
                await asyncio.sleep(self.sensing_interval)
            except SensorFatalError:
                logger.critical(f'Sensor cannot continue and will be killed! Reason: {e}')
                return
            except Exception as e:
                logger.error(f'Error while sensing! Details: {e}')

    def start(self):
        """
        Prepares the sensor and starts it.

        The basic implementation calls the sense method in a loop, it will
        stop when the state of the sensor changes to SensorState.STOPPED

        This is a basic implementation and can be overridden if necessary.
        """
        self.state = SensorState.RUNNING
        self._start_task = asyncio.get_event_loop().create_task(self._start())

    async def stop(self):
        """
        Stops the sensor main loop and release resources.

        This is a basic implementation and can be overridden if necessary.
        """
        self.state = SensorState.STOPPED

    def status(self):
        return Sensor(
            id=self.id, template=self.sensor_template, status=self._state,
            events_count=self._events_counter, events_forwarded=self._events_forwarded,
            events_discarded=self._events_discarded
        )

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        self._state = state


class DummySensor(AbstractSensor):
    """
    Dummy sensor that emits io.orchd.events.system.Test events.
    """
    def __init__(self, sensor_template,  communicator: AbstractCommunicator):
        super().__init__(sensor_template, communicator)
        self.state = SensorState.READY

    template = SensorTemplate(
        name='io.orchd.sensor_template.DummySensor',
        description='A dummy Sensor to be used for testing purposes',
        version='1.0',
        sensor='orchd_sdk.sensor.DummySensor',
        communicator='orchd_sdk.sensor.LocalCommunicator',
        parameters={'some': 'data'},
        sensing_interval=0
    )

    async def sense(self):
        await asyncio.sleep(1)
        await self.communicator.emit_event(
            Event(event_name='io.orchd.events.system.Test', data={'dummy': 'data'})
        )


class LocalCommunicator(AbstractCommunicator):
    """
    Object Message Communicator for emitting events.

    This communicator emits events by invoking the given ReactionsEventBus
    that is used by the Reactor. This communicator do not requires authentication
    since it is part of the system. It is intended to provide more performance.

    If no ReactionsBusEvent object is given the default one is used instead.

    It is recommended that this communicator to be Ued by trusted sensors.
    """

    def __init__(self, event_bus: ReactionsEventBus = None):
        super().__init__()
        self.event_bus = event_bus or global_reactions_event_bus

    async def emit_event(self, event: Event):
        """
        Emits an event using the global ReactionsEventBus
        :param event: Event to emit.
        """
        self.event_bus.event(event)

    async def authenticate(self):
        """
        no-op since local communicator do not need to authenticate.
        """
        pass

    def close(self):
        """
        no-op since it is a local communicator and do not uses additions
        resources/connections.
        """



from typing import List

from pydantic import TypeAdapter

from orchd_sdk.models import ReactionInfo, ReactionTemplate, Ref


class ReactionClient:

    def __init__(self, orchd_client):
        self.orchd_client = orchd_client

    async def get_reactions(self) -> List[ReactionInfo]:
        reactions = await self.orchd_client.get('/reactions')
        adapter = TypeAdapter(List[ReactionInfo])
        return adapter.validate_python(reactions)

    async def get_reaction(self, reaction_id: str) -> ReactionInfo:
        reaction = await self.orchd_client.get(f'/reactions/{reaction_id}')
        return ReactionInfo(**reaction)

    async def add_reaction(self, template_id: str) -> ReactionInfo:
        response = await self.orchd_client.post('/reactions', Ref(id=template_id).model_dump())
        return ReactionInfo(**response)

    async def remove_reaction(self, reaction_id: str) -> str:
        response = await self.orchd_client.delete(f'/reactions/{reaction_id}')
        return response

    async def get_reaction_templates(self) -> List[ReactionTemplate]:
        templates = await self.orchd_client.get('/reactions/templates/')
        adapter = TypeAdapter(List[ReactionTemplate])
        return adapter.validate_python(templates)

    async def get_reaction_template(self, template_id: str) -> ReactionTemplate:
        template = await self.orchd_client.get(f'/reactions/templates/{template_id}')
        return ReactionTemplate(**template)

    async def add_reaction_template(self, template: ReactionTemplate) -> ReactionTemplate:
        response = await self.orchd_client.post('/reactions/templates/', template.model_dump())
        return ReactionTemplate(**response)

    async def remove_reaction_template(self, template_id: str) -> str:
        response = await self.orchd_client.delete(f'/reactions/templates/{template_id}/')
        return response

    async def add_sink_to_reaction(self, reaction_id: str, template_id: str) -> str:
        response = await self.orchd_client.post(f'/reactions/{reaction_id}/sinks',
                                                Ref(id=template_id).model_dump())
        return response

    async def remove_sink_from_reaction(self, reaction_id: str, sink_id: str) -> str:
        response = await self.orchd_client.delete(f'/reactions/{reaction_id}/sinks/{sink_id}')
        return response
